/**
 *
 * MDEditor - A jQuery Markdown Editor Plugin
 * https://github.com/ShinyChang/MarkDownEditor
 *
 * Copyright (c) 2014, ZhiHong Chang, voluntariness
 *
 * Licensed under the MIT license:
 * http://www.opensource.org/licenses/MIT
 */

(function(factory) {
    'use strict';
    if (typeof define === 'function' && define.amd) {
        // Register as an anonymous AMD module:
        define([
            'jquery',
            'jquery.autosize'
        ], factory);
    } else {
        // Browser globals:
        factory(window.jQuery);
    }
}(function($) {
    var Helper = {
        requestFullscreen: function(element) {
            if (element.requestFullscreen) {
                element.requestFullscreen();
            } else if (element.mozRequestFullScreen) {
                element.mozRequestFullScreen();
            } else if (element.webkitRequestFullscreen) {
                element.webkitRequestFullscreen();
            } else if (element.msRequestFullscreen) {
                element.msRequestFullscreen();
            }
        },
        exitFullscreen: function() {
            if (document.exitFullscreen) {
                document.exitFullscreen();
            } else if (document.mozCancelFullScreen) {
                document.mozCancelFullScreen();
            } else if (document.webkitExitFullscreen) {
                document.webkitExitFullscreen();
            } else if (document.msExitFullscreen) {
                document.msExitFullscreen();
            }
        }
    };

    var MDEditor = function(element, options) {

        /*
         * list: 記錄陣列
         * idx: 讀/寫的指定位置
         * max: 陣列最大長度
         * last: KeyIn的最後 keyCode
         */
        this.records = {
            list: [],
            idx: 0,
            max: 1000,
            last: null
        };
        this.options = $.extend(true, {}, $.fn.MDEditor.defaults, options);
        this.$element = $(element);

        this.$wrapper = this.$element.wrap("<div class='mdeditor'></div>").parent();
        this.$preview = $("<div>").attr({
            'class': 'mdeditor-preview'
        }).appendTo(this.$wrapper);
        this.$wrapper.append(
            "<div class='mdeditor-toolbar'>\
                <div class='mdeditor-btn mdeditor-btn-preview' data-role='preview'>Preview</div>\
                <div class='mdeditor-btn mdeditor-btn-fullscreen' data-role='fullscreen'>Fullscreen</div>\
            </div>");

        this.parser = this.options.parser;

        if (this.options.autosize) {
            this.$element.autosize();
        }

        this.bindEvent();
        this.renderPreview();
        this.saveRecord();
    }

    MDEditor.prototype = {
        constructor: MDEditor,

        /*
         * Realtime parse the context
         */
        bindEvent: function() {
            this.$element.bind("input.MDEditor", $.proxy(function() {
                this.renderPreview();
                this.$element.trigger('update', [{
                    html: this.getHTML(),
                    md: this.getMD()
                }]);
            }, this));

            this.$element.bind("cut.MDEditor", $.proxy(function(e) {
                this.saveRecord();
            }, this));
            this.$element.bind("paste.MDEditor", $.proxy(function(e) {
                this.saveRecord();
            }, this));

            this.$element.bind("keydown.MDEditor", $.proxy(function(e) {

                switch (true) {
                    case (e.keyCode === 27): // Esc
                        return false; // avoid blur

                    case (e.keyCode === 9): // Tab
                        if (e.ctrlKey) { // tab switch (browser default)
                            return;
                        }
                        e.preventDefault();
                        var lines = this.getLinesText(),
                            line_start = e.target.selectionStart,
                            line_end = e.target.selectionEnd;
                        this.saveRecord();
                        if (e.target.selectionStart == e.target.selectionEnd && e.shiftKey) {

                            var len = lines[0].text.match(/^[ ]*/).toString().length;

                            // 游標前面都是空白時
                            if (line_start == lines[0].start + len) {
                                len = len > 4 ? 4 : len;
                                line_start = (line_end -= len);
                                this.insertText('', lines[0].start, lines[0].start + len);
                            } else {
                                var sub = 4 - (lines[0].text.length % 4);
                                line_start += sub;
                                line_end += sub;
                                this.insertText(this.copyStr(' ', sub), e.target.selectionStart);
                            }

                        } else if (e.shiftKey) {

                            line_start -= lines[0].text.match(/^[ ]{0,4}/).toString().length;
                            var slen = 0,
                                str = [];
                            for (var i in lines) {
                                slen = lines[i].text.match(/^[ ]{0,4}/).toString().length;
                                line_end -= slen;
                                str.push(lines[i].text.substr(slen));
                            }
                            this.insertText(str.join('\n'), lines[0].start, lines[i].end);

                        } else { // 多行縮排時
                            line_start += 4;
                            line_end += lines.length * 4;
                            var str = [];
                            for (var i in lines) {
                                str.push('    ' + lines[i].text);
                            }
                            this.insertText(str.join('\n'), lines[0].start, lines[i].end);
                        }
                        e.target.setSelectionRange(line_start, line_end);
                        break;

                    case (e.keyCode === 36): // Home
                        e.preventDefault();

                        var lines = this.getLinesText(),
                            sidx = lines[0].text.match(/^[ ]*/).toString().length + lines[0].start;
                        sidx = sidx == e.target.selectionStart ? lines[0].start : sidx;
                        var eidx = e.shiftKey ? e.target.selectionEnd : sidx;
                        e.target.setSelectionRange(Math.min(sidx, eidx), Math.max(sidx, eidx));
                        return;

                    case (e.keyCode === 89 && e.ctrlKey): // y
                        e.preventDefault();
                        this.nextRecord();
                        break;

                    case (e.keyCode === 90 && e.ctrlKey): // z
                        e.preventDefault();
                        this.prevRecord();
                        break;

                        // BackSpace Enter Space
                    case ([8, 13, 32].indexOf(e.keyCode) !== -1):
                        if (e.keyCode != this.records.last) {
                            this.saveRecord();
                            this.records.last = e.keyCode;
                        }
                        return;

                        // Left, Up, Right, Down, Shift, Ctrl, Alt, Caps Lock, Menu
                    case ([37, 38, 39, 40, 16, 17, 18, 20, 93].indexOf(e.keyCode) !== -1):
                        // F1 ~ F12
                    case (e.keyCode >= 112 && e.keyCode <= 123):
                        // 以上按鍵不觸發 saveRecord
                        return;

                        // 記錄最後一次的 keyCode , 若為連續 space, enter,backspace 不記錄
                    default:
                        this.records.last = e.keyCode;
                        return;
                }
                this.renderPreview();
                this.$element.trigger('update', [{
                    html: this.getHTML(),
                    md: this.getMD()
                }]);
            }, this));

            this.$preview.on('click.MDEditor', 'a', function(e) {
                e.preventDefault();
                alert('Link to: ' + $(this).attr('href'));
            });

            this.$wrapper.on('click.MDEditor', "[data-role='preview']", $.proxy(this.togglePreview, this));
            this.$wrapper.on('click.MDEditor', "[data-role='fullscreen']", $.proxy(this.toggleFullscreen, this));
        },
        getHTML: function() {
            return this.parser(this.getMD());
        },
        getMD: function() {
            return this.$element.val();
        },
        renderPreview: function() {
            this.$preview.html(this.getHTML());
        },
        enterPreview: function() {
            this.$wrapper.addClass('preview');
        },
        exitPreview: function() {
            this.$wrapper.removeClass('preview');
        },
        togglePreview: function(flag) {
            this.$wrapper.toggleClass('preview');
        },
        enterFullscreen: function() {
            this.exitPreview(); // exit preview first
            this.$element.trigger('autosize.destroy');
            this.$wrapper.addClass('fullscreen')
            Helper.requestFullscreen(this.$wrapper.get(0));
            $(document).bind('keyup.MDEditor', $.proxy(function(e) {
                if (e.keyCode === 27 || e.keyCode === 122) {
                    if (this.$element.is(":focus")) {
                        return;
                    }
                    this.exitFullscreen();
                }
            }, this));
        },
        exitFullscreen: function() {
            this.exitPreview();
            this.$element.removeAttr('style').autosize();
            this.$wrapper.removeClass('fullscreen');
            Helper.exitFullscreen();
            $(document).unbind('keyup.MDEditor');
        },
        toggleFullscreen: function() {
            this.$wrapper.hasClass('fullscreen') ? this.exitFullscreen() : this.enterFullscreen();
        },

        saveRecord: function() {
            if (this.records.list.length > this.records.idx) {
                this.records.list = this.records.list.splice(0, this.records.idx);
            }
            if (this.records.idx + 1 >= this.records.max) {
                this.records.list.shift();
                this.records.list[this.records.idx] = {
                    text: this.$element.val(),
                    start: this.$element[0].selectionStart,
                    end: this.$element[0].selectionEnd
                };
            } else {
                this.records.list[this.records.idx] = {
                    text: this.$element.val(),
                    start: this.$element[0].selectionStart,
                    end: this.$element[0].selectionEnd
                };
                this.records.idx++;
            }
        },
        prevRecord: function() {
            if (this.records.idx > 0) {

                if (this.records.idx == this.records.list.length) {
                    this.saveRecord();
                    this.records.idx--;
                }

                this.records.idx--;
                record = this.records.list[this.records.idx];
                this.$element.val(record.text);
                this.$element[0].setSelectionRange(record.start, record.end);
            }
        },
        nextRecord: function() {

            if (this.records.idx < this.records.list.length - 1) {
                this.records.idx++;
                record = this.records.list[this.records.idx];
                this.$element.val(record.text);
                this.$element[0].setSelectionRange(record.start, record.end);
            }
        },

        // 取得 TextArea 內選取的文字各行
        getLinesText: function() {

            var lines = [],
                line_start = 0,
                line_end = 0;

            for (var i in (list = this.$element.val().split('\n'))) {

                line_end = line_start + list[i].length;

                if (this.$element[0].selectionStart <= line_end) {
                    lines.push({
                        text: list[i],
                        start: line_start,
                        end: line_end
                    });
                }

                if (this.$element[0].selectionEnd <= line_end) {
                    break;
                }

                line_start = line_end + 1; // 換行
            }

            return lines;
        },

        // 插入文字
        insertText: function(txt, start, end) {
            start = start || 0;
            end = end || start;
            var tmp = this.$element.val().substr(end);
            this.$element.val(this.$element.val().substr(0, start) + txt + tmp);
        },
        copyStr: function(str, len) {
            return (new Array(len + 1)).join(str);
        }
    }

    var old = $.fn.MDEditor;

    /*
     * MDEDITOR PLUGIN DEFINITION
     */
    $.fn.MDEditor = function(options, value) {
        return this.each(function() {
            var $this = $(this),
                instance = $this.data('MDEditor'),
                opt = typeof options === 'object' && options;

            if (!instance) {
                $this.data('MDEditor', (instance = new MDEditor(this, opt)));
            }

            if (typeof options === 'string' && typeof instance[options] === 'function') {
                instance[options](value);
            }
        });
    }


    /*
     * MDEDITOR DEFAULTS
     */
    $.fn.MDEditor.defaults = {
        parser: undefined,
        autosize: true // this require jQuery autosize plugin
    }


    /*
     * MDEDITOR CONSTRUCTOR
     */
    $.fn.MDEditor.Constructor = MDEditor;


    /*
     * MDEDITOR NO CONFLICT
     */
    $.fn.MDEditor.noConflict = function() {
        $.fn.MDEditor = old;
        return this;
    }
}));
